#!/usr/bin/perl -w

use strict;
use URI::Escape;
use Data::Dumper;

my $prec=2;

my $nosizecheck;
if( defined($ARGV[0]) && $ARGV[0] eq "-d" ) {
  $nosizecheck= 1;
  shift @ARGV;
}

if( @ARGV != 1 || $ARGV[0] =~ /^-h$|^-?-help$/i ) {
  print STDERR <<EOF;
usage: texlinkcmp [-d] <glob pattern>
For an every-vs-every cross comparison, the glob pattern must contain a `#' for
the independent variable.  For multiple pairwise comparisons, a `#' also
denotes the comparison criterion, and an `\@' is required to indicate the
property for which separate comparisons are to be made.
Don't forget to escape the glob pattern from the shell.
-d disables the check that corresponding files contain the same number of URLs.
Use it when comparing or averaging comparisons between lists with different
sizes.
EOF
  exit 1;
}

###########################################################################
#	Obtain file names, read files and check consistency
###########################################################################

# We now build the glob pattern and two regular expressions, one to filter the
# glob results and one to generate a hash key representing the independent
# variables later.  Strictly speaking, the argument to this script is not a
# glob pattern, as we interpret wildcards to match only word characters.
my $pat= $ARGV[0];
my $compare= "^" . quotemeta($pat) . "\$";
$pat =~ tr/#@/**/;
$compare =~ s/\\\*/\\w\*/g;
$compare =~ s/\\\?/\\w/g;
my $re= $compare;
my $classify;
if( $compare =~ /\@/ ) {	# multiple pairwise comparisons
  $classify= $compare;
  $classify =~ s/\\\@/\(\\w\*\)/g;
  $classify =~ s/\\\#/\\w\*/g;
}
else { 		# large triangluar any-with-any comparison
  $classify= "";
}
$compare =~ s/\\\#/\(\\w\*\)/g;
$compare =~ s/\\\@/\\w\*/g;
$re =~ s/\\[\#\@]/\\w\*/g;

# The sorting ensures that we will later compare corresponding URL lists:
my @files= sort grep $_ =~ /$re/, glob($pat);
die "No files matching the glob pattern found" unless @files;
my %sorted;

# Read URL lists.  Because some of the URLs may be partly escaped (according to
# RFC 2396), we unescape them.
for my $file (@files) {
  local $/= undef;
  open FIL, "<$file" or die "Can't open $file for reading";
  my $content= <FIL>;
  close FIL;
  my $key= join "-", $file =~ /$compare/;
  if( $classify ) {
    my $ckey= join "-", $file =~ /$classify/;
    push @{$sorted{$ckey}->{$key}}, [ map uri_unescape($_), split /\n/, $content ];
  }
  else {
    push @{$sorted{$key}}, [ map uri_unescape($_), split /\n/, $content ];
  }
}

my $maxind;

if( $classify ) {
  for my $key (keys(%sorted)) {
    die "Cannot only compare two cases in classified comparison"
    	if scalar(keys(%{$sorted{$key}})) != 2;
    my ($k1, $k2)= keys %{$sorted{$key}};
    die "Found a different number of corresponding files for the independent" .
    " variable $key" unless @{$sorted{$key}->{$k1}} == @{$sorted{$key}->{$k2}};
    for my $ind (0..$#{$sorted{$key}->{$k1}}) {
      die "Some corresponding files for the independent variable $key have " .
      	"a different number of lines"
     unless @{$sorted{$key}->{$k1}->[$ind]} == @{$sorted{$key}->{$k2}->[$ind]} || $nosizecheck;
    }
  }

  print STDERR "Found ", scalar(keys %sorted), " classes to compare files in.\n";
}
else {
  my ($min, $max)= (0x7fff, 0);
  for my $key (keys(%sorted)) {
    $min= @{$sorted{$key}} if @{$sorted{$key}} < $min;
    $max= @{$sorted{$key}} if @{$sorted{$key}} > $max;
  }
  die "Not all values of the independent variable have the same number of files ($min to $max)" unless $min == $max;

  my ($allminsize, $allmaxsize)= (0x7fff, 0);
  $maxind= $min - 1;
  for my $ind (0..$maxind) {
    my ($minsize, $maxsize)= (0x7fff, 0);
    for my $list (map $_->[$ind], values(%sorted)) {
      $minsize= @$list if @$list < $minsize;
      $maxsize= @$list if @$list > $maxsize;
    }
    die "Not all files have the same number of lines ($minsize to $maxsize) for some values of the averaging variables" unless $minsize == $maxsize || $nosizecheck;
    $allminsize= $minsize if $minsize < $allminsize;
    $allmaxsize= $maxsize if $maxsize > $allmaxsize;
  }

  print STDERR "Found $min files for each of ", scalar(keys(%sorted)),
	  " values of the independent variable, with ",
	  ($allminsize == $allmaxsize? $allmaxsize :
	  "between $allminsize and $allmaxsize"), " lines per file.\n";
}

###########################################################################
#			Compare URL lists
###########################################################################

sub max2 { $_[0] > $_[1] ? $_[0] : $_[1]; }

# Compare list of URLs
# -> Two references to arrays of URLs
# <- Reference to array containing three numbers indicating the differences
#    between the URL lists:
#  - the fraction of URLs in the first array not found in the second
#  - the number of inversions in the permutation of the URLs which were found
#    in both arrays, normalised with the maximum possible number of inversions,
#    taking into account the number of URLs not found
#  - the average difference in the rankings of an URL in the two arrays divided
#    by the maximum possible average distance

sub cmpllist
{
  my ($l1, $l2)= @_[0..1];
  my ($avgrank1, $avgrank2, $avgoff)= (0) x 3;
  my $totalnot= 0;
  my (@ranks1, @ranks2);

  if( $#$l1 > $#$l2 ) {
    my $tmp= $l1;
    $l1= $l2;
    $l2= $tmp;
  }
  for my $ind1 (0..$#$l1) {
    my @found;
    for my $ind2 (0..$#$l2) {
      push @found, $ind2 if $$l1[$ind1] eq $$l2[$ind2];
    }
    if( !@found ) {	# not found
      ++$totalnot;
    }
    else {
      my $newind= shift @found;
      if( @found ) {		# multiple occurrences in second list
	for my $f (@found) {
	  $newind= $f if abs($f - $ind1) < abs($newind - $ind1);
	}
	# If the URL occurs several times in both lists, choosing the nearest
	# occurrence in both cases may assign the same occurrence in the second
	# list to both in the first, which makes no sense.  This should not
	# happen in the lists we compare here, so warn the user.
	print STDERR $$l1[$ind1], " found more than once, minimum distance ",
		$newind - $ind1, "\n";
      }
      push @ranks1, $ind1;
      push @ranks2, $newind;
    }
  }

  # Average number of entries an URL has been moved in the second array
  # relative to the first
  @ranks2= sort {$a <=> $b} @ranks2;
  for my $ind (0..$#$ranks1) {
    $avgoff += abs($ranks1[$ind] - $ranks2[$ind]);
  }
  $svgoff /= @ranks1 if @ranks1;
  # Normalise by division by maximum offset
  $avgoff /= max2(scalar(@$l1), scalar(@$l2)) - @ranks1
  	if @ranks1 < max2(scalar(@$l1), scalar(@$l2));

  # Number of inversions in the permutation of URL order:
  my $inv= 0;
  if( @ranks1 > 1 ) {
    for my $p1 (0..$#ranks1) {
      for my $p2 (0..$p1-1) {
	++$inv if $ranks2[$p2] > $ranks2[$p1];
      }
    }
    $inv /= @ranks1 * (@ranks1 - 1) / 2;
  }
  $totalnot /= @$l1;
  return ($totalnot, $inv, $avgoff);
}

# Order of the values of the independent variable in the table columns:
my @keyorder= sort keys %sorted;
# Turn hash keys (themselves derived from file names) into table column/row
# headings and sort them according to heading
my %headings= map {; $_ => $_ } @keyorder;
if( grep(/\b(?:ie|fi|op)\b/, @keyorder) > 2 ) {	# browser
  map { s/(\w+)/\u$1/g; s/\bie\b/IE/i; s/-/ \/ /g; } values %headings;
}
elsif( grep(/^[wl]$/, @keyorder) == 2 ) {	# OS
  map { s/w/Windows/; s/l/Linux/; } values %headings;
}
elsif( grep(/\b(?:de|fr|ch)\b/, @keyorder) > 2 ) {  # country domain
  map { s/^/\./; } values %headings;
}
elsif( grep(/^(?:argv|las|wash)/, @keyorder) > 2 ) {	# search term
  map { s/^(?:aco|auto).*/Autocorrelation/i; s/^argv.*/ARGV/i;
  	s/^hard.*/Hard disk benchmark/i; s/^laser.*/Laser/i;
	s/^shel.*/Shelley/i; s/^was.*/Washington/i; } values %headings;
}
elsif( grep(length($_) > 5, @keyorder) > 2 ) {		# client country
  map { s/.*bruchs.*/D/i; s/.*anderm.*/CH (de)/i; s/.*colma.*/F/i;
  	s/.*innsbr.*/A/i; s/.*lugan.*/CH (it)/i;  s/.*como.*/I/i; }
	values %headings;
}
@keyorder= sort { $headings{$a} cmp $headings{$b} } @keyorder;

# 2D results array
my %avgresult;

if( $classify ) {
  for my $key (@keyorder) {
    my ($k1, $k2)= keys %{$sorted{$key}};
    for my $ind (0..$#{$sorted{$key}->{$k1}}) {
      my @cmp= cmpllist($sorted{$key}->{$k1}->[$ind], $sorted{$key}->{$k2}->[$ind]);
      for (0..2) { $avgresult{$key}->[$_] += $cmp[$_]; }
    }
    for (0..2) { $avgresult{$key}->[$_] /= @{$sorted{$key}->{$k1}}; }
  }
}
else {
  for my $ind (0..$maxind) {
    for my $keyy (reverse @keyorder) {
      for my $keyx (@keyorder) {
	my @cmp= cmpllist($sorted{$keyx}->[$ind], $sorted{$keyy}->[$ind]);
	for (0..2) { $avgresult{$keyy}->{$keyx}->[$_] += $cmp[$_]; }
	last if $keyx eq $keyy;
      }
    }
  }

  for my $keyy (reverse @keyorder) {
    for my $keyx (@keyorder) {
      for (0..2) { $avgresult{$keyy}->{$keyx}->[$_] /= $maxind+1; }
      last if $keyx eq $keyy;
    }
  }
}

# print Dumper(\%avgresult);

###########################################################################
#			TeX Output
###########################################################################

if( $classify ) {	# pair comparison
  print <<'EOF';
\def\pad{\ \ }
\hbox to \hsize{\hfill\vbox{\offinterlineskip\halign{
\vfat height2.75ex depth1.25ex \pad # \hfil\pad\vfat & 
\pad # \hfil&\pad # \hfil&\pad # \hfil\pad\vfat\cr
\expandafter\ifx\csname linkcmptitle\endcsname\relax\else
\hfat
Search term & \omit\span\omit\span\hfil\linkcmptitle\cr\fi
EOF
  my $prevline= "\\hfat\n";
  for my $key (@keyorder) {
    print $prevline, $headings{$key}, " & ",
    	join(" & ", map(sprintf("%.2f", $_), @{$avgresult{$key}})), "\\cr\n";
    $prevline= "\\hthin\n";
  }
  print "\\hfat}}\\hfill}\n";
}
else {			# every-vs-every comparison
  print <<'EOF';
\def\pad{\ \ }
\def\hcellrule{\omit\leaders\hrule height 0.3pt\hfil}
\def\h3cellrule{\omit\span\omit\span\omit\leaders\hrule height 0.3pt\hfil}
\def\hfcellrule{\omit\leaders\hrule height 0.6pt\hfil}
\hbox to \hsize{\hfill\vbox{\offinterlineskip\halign{
\vfat height2.75ex depth1.25ex \pad # \hfil\pad\vfat
EOF
  my $celltempl= "\\pad \# \\hfil&\\pad \# \\hfil&\\pad \# \\hfil\\pad";
  print "&$celltempl\\vthin\n" x ($#keyorder-1), "&$celltempl\\vfat\\cr\n"; 
  print "\\hfat\n";
  print "\\expandafter\\ifx\\csname linkcmptitle\\endcsname\\relax\\else\\linkcmptitle\\fi\n";
  for (map $headings{$_}, @keyorder[0..$#keyorder-1]) {
    print "&\\omit\\span\\omit\\span $_\n";
  }
  print "\\cr\\hfat\n";

  my $prevlnend= "";
  my $lastkeyind= $#keyorder-1;
  for my $keyy ((reverse @keyorder)[0..$#keyorder-1]) {
    print $prevlnend, $headings{$keyy}, "\n";
    for my $keyx (@keyorder[0..$lastkeyind-1]) {
      print "  & ", join(" & ", map(sprintf("%.${prec}f", $_), @{$avgresult{$keyy}->{$keyx}})), "\n";
    }
    my $lastkeyx= $keyorder[$lastkeyind];
    print "  & ", join(" & ", map(sprintf("%.${prec}f", $_), @{$avgresult{$keyy}->{$lastkeyx}}[0..1])), "\n";
    print "   & \\omit\\pad ", sprintf("%.${prec}f", $avgresult{$keyy}->{$lastkeyx}->[2]),
	  " \\hfil\\pad\\vfat\n";
    $prevlnend= "\\cr\n\\hcellrule";
    for my $keyx (@keyorder[0..$lastkeyind-1]) {
      $prevlnend .= "& \\h3cellrule";
    }
    $prevlnend .= "& \\hfcellrule" x 3 . "\\cr\n";
    --$lastkeyind;
  }
  print "\\cr\n\\hfcellrule", " &\\hfcellrule" x 3, " \\cr\n}}\\hfill}\n";
}

